"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = _default;

var _generator = _interopRequireDefault(require("@babel/generator"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread(); }

function _nonIterableSpread() { throw new TypeError("Invalid attempt to spread non-iterable instance"); }

function _iterableToArray(iter) { if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter); }

function _arrayWithoutHoles(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } }

/**
 * # Design By Contract Transformer
 */
function _default(_ref) {
  var t = _ref.types,
      template = _ref.template;
  var defaultNames = {
    assert: 'assert',
    precondition: 'pre',
    postcondition: 'post',
    invariant: 'invariant',
    return: 'it',
    old: 'old'
  };
  var defaultEnvNames = {
    development: "dev",
    test: "test",
    production: "prod"
  };
  var NAMES = Object.assign({}, defaultNames);
  var ENV_NAMES = Object.assign({}, defaultEnvNames);
  var ENV_LABEL_SET;
  var guard = template("\n    if (!CONDITION) {\n      throw new Error(MESSAGE);\n    }\n  ");
  var guardFn = template("\n    const ID = (IT) => {\n      CONDITIONS;\n      return IT;\n    }\n  ");

  function assemblePrecondition(path) {
    var body = path.get('body');
    var fn = path.getFunctionParent();
    var name = fn.node.id ? "\"".concat(fn.node.id.name, "\" ") : ' ';

    if (body.isExpressionStatement()) {
      var condition = body.get('expression');
      var message;

      if (condition.isSequenceExpression()) {
        var expressions = condition.get('expressions');
        condition = expressions[0];
        message = expressions[1].node;
      } else {
        message = t.stringLiteral("Function ".concat(name, "precondition failed: ").concat((0, _generator.default)(condition.node).code));
      }

      path.replaceWith(guard({
        CONDITION: staticCheck(condition),
        MESSAGE: message
      }));
      return;
    }

    body.traverse({
      "VariableDeclaration|Function|AssignmentExpression|UpdateExpression|YieldExpression|ReturnStatement|AwaitExpression": function VariableDeclarationFunctionAssignmentExpressionUpdateExpressionYieldExpressionReturnStatementAwaitExpression(item) {
        throw path.buildCodeFrameError("Preconditions cannot have side effects.");
      },
      ExpressionStatement: function ExpressionStatement(statement) {
        var condition = statement.get('expression');
        var message;

        if (condition.isSequenceExpression()) {
          var _expressions = condition.get('expressions');

          condition = _expressions[0];
          message = _expressions[1].node;
        } else {
          message = t.stringLiteral("Function ".concat(name, "precondition failed: ").concat((0, _generator.default)(condition.node).code));
        }

        statement.replaceWith(guard({
          CONDITION: staticCheck(condition),
          MESSAGE: message
        }));
      }
    });

    if (body.isBlockStatement()) {
      path.replaceWithMultiple(path.get('body').node.body);
    } else {
      path.replaceWith(path.get('body'));
    }
  }

  function assemblePostcondition(path) {
    var body = path.get('body');
    var fn = path.getFunctionParent();
    var name = fn.node.id ? "\"".concat(fn.node.id.name, "\" ") : ' ';
    var conditions = [];

    if (body.isExpressionStatement()) {
      var condition = body.get('expression');
      var message;

      if (condition.isSequenceExpression()) {
        var expressions = condition.get('expressions');
        condition = expressions[0];
        message = expressions[1].node;
      } else {
        message = t.stringLiteral("Function ".concat(name, "postcondition failed: ").concat((0, _generator.default)(condition.node).code));
      }

      conditions.push(guard({
        CONDITION: staticCheck(condition),
        MESSAGE: message
      }));
    } else {
      body.traverse({
        "VariableDeclaration|Function|AssignmentExpression|UpdateExpression|YieldExpression|ReturnStatement|AwaitExpression": function VariableDeclarationFunctionAssignmentExpressionUpdateExpressionYieldExpressionReturnStatementAwaitExpression(item) {
          throw path.buildCodeFrameError("Postconditions cannot have side effects.");
        },
        CallExpression: function CallExpression(call) {
          var callee = call.get('callee');
          var args = call.get('arguments');

          if (!callee.isIdentifier() || callee.node.name !== NAMES.old || call.scope.hasBinding(NAMES.old) || args.length === 0) {
            return;
          }

          var argument = args[0];
          var id = call.scope.generateUidIdentifierBasedOnNode(argument.node);
          fn.scope.push({
            id: id,
            init: argument.node,
            kind: 'const'
          });
          call.replaceWith(id);
        },
        ExpressionStatement: function ExpressionStatement(statement) {
          var condition = statement.get('expression');
          var message;

          if (condition.isSequenceExpression()) {
            var _expressions2 = condition.get('expressions');

            condition = _expressions2[0];
            message = _expressions2[1].node;
          } else {
            message = t.stringLiteral("Function ".concat(name, "postcondition failed: ").concat((0, _generator.default)(condition.node).code));
          }

          statement.replaceWith(guard({
            CONDITION: staticCheck(condition),
            MESSAGE: message
          }));
        }
      });
      conditions.push.apply(conditions, _toConsumableArray(body.node.body));
    }

    var id = path.scope.generateUidIdentifier("".concat(fn.node.id ? fn.node.id.name : 'check', "Postcondition"));
    fn.get('body').get('body')[0].insertBefore(guardFn({
      ID: id,
      CONDITIONS: conditions,
      IT: t.identifier(NAMES.return)
    }));
    path.remove();
    return id;
  }

  function assembleAssertion(path) {
    var body = path.get('body');
    var fn = path.getFunctionParent();
    var name = fn && fn.node && fn.node.id ? "\"".concat(fn.node.id.name, "\"") : '';

    if (body.isExpressionStatement()) {
      var condition = body.get('expression');
      var message;

      if (condition.isSequenceExpression()) {
        var expressions = condition.get('expressions');
        condition = expressions[0];
        message = expressions[1].node;
      } else if (name) {
        message = t.stringLiteral("Function ".concat(name, " assertion failed: ").concat((0, _generator.default)(condition.node).code));
      } else {
        message = t.stringLiteral("Assertion failed: ".concat((0, _generator.default)(condition.node).code));
      }

      path.replaceWith(guard({
        CONDITION: staticCheck(condition),
        MESSAGE: message
      }));
      return;
    }

    body.traverse({
      "VariableDeclaration|Function|AssignmentExpression|UpdateExpression|YieldExpression|ReturnStatement|AwaitExpression": function VariableDeclarationFunctionAssignmentExpressionUpdateExpressionYieldExpressionReturnStatementAwaitExpression(item) {
        throw path.buildCodeFrameError("Assertions cannot have side effects.");
      },
      ExpressionStatement: function ExpressionStatement(statement) {
        var condition = statement.get('expression');
        var message;

        if (condition.isSequenceExpression()) {
          var _expressions3 = condition.get('expressions');

          condition = _expressions3[0];
          message = _expressions3[1].node;
        } else if (name) {
          message = t.stringLiteral("Function ".concat(name, " assertion failed: ").concat((0, _generator.default)(condition.node).code));
        } else {
          message = t.stringLiteral("Assertion failed: ".concat((0, _generator.default)(condition.node).code));
        }

        statement.replaceWith(guard({
          CONDITION: staticCheck(condition),
          MESSAGE: message
        }));
      }
    });

    if (body.isBlockStatement()) {
      path.replaceWithMultiple(path.get('body').node.body);
    } else {
      path.replaceWith(path.get('body'));
    }
  }

  function assembleInvariant(path) {
    var body = path.get('body');
    var fn = path.getFunctionParent();
    var name = fn.node.id ? "\"".concat(fn.node.id.name, "\" ") : ' ';
    var conditions = [];

    if (body.isExpressionStatement()) {
      var condition = body.get('expression');
      var message;

      if (condition.isSequenceExpression()) {
        var expressions = condition.get('expressions');
        condition = expressions[0];
        message = expressions[1].node;
      } else {
        message = t.stringLiteral("Function ".concat(name, "invariant failed: ").concat((0, _generator.default)(condition.node).code));
      }

      conditions.push(guard({
        CONDITION: staticCheck(condition),
        MESSAGE: message
      }));
    } else {
      body.traverse({
        "VariableDeclaration|Function|AssignmentExpression|UpdateExpression|YieldExpression|ReturnStatement|AwaitExpression": function VariableDeclarationFunctionAssignmentExpressionUpdateExpressionYieldExpressionReturnStatementAwaitExpression(item) {
          throw path.buildCodeFrameError("Invariants cannot have side effects.");
        },
        ExpressionStatement: function ExpressionStatement(statement) {
          var condition = statement.get('expression');
          var message;

          if (condition.isSequenceExpression()) {
            var _expressions4 = condition.get('expressions');

            condition = _expressions4[0];
            message = _expressions4[1].node;
          } else {
            message = t.stringLiteral("Function ".concat(name, "invariant failed: ").concat((0, _generator.default)(condition.node).code));
          }

          statement.replaceWith(guard({
            CONDITION: staticCheck(condition),
            MESSAGE: message
          }));
        }
      });
      conditions.push.apply(conditions, _toConsumableArray(body.node.body));
    }

    var id = path.scope.generateUidIdentifier("".concat(fn.node.id ? fn.node.id.name : 'check', "Invariant"));
    path.parentPath.get('body')[0].insertBefore(guardFn({
      ID: id,
      CONDITIONS: conditions,
      IT: t.identifier(NAMES.return)
    }));
    path.remove();
    return id;
  }

  function extractEnvironmentCode(path) {
    var body = path.get('body');

    if (body.isBlockStatement()) {
      path.replaceWithMultiple(path.get('body').node.body);
    } else {
      path.replaceWith(path.get('body'));
    }
  }

  function expression(input) {
    var fn = template(input);
    return function () {
      var node = fn.apply(void 0, arguments);
      return getExpression(node);
    };
  }

  function staticCheck(expression) {
    var _expression$evaluate = expression.evaluate(),
        confident = _expression$evaluate.confident,
        value = _expression$evaluate.value;

    if (confident && !value) {
      throw expression.buildCodeFrameError("Contract always fails.");
    }

    return expression.node;
  }

  return {
    visitor: {
      Program: function Program(path, _ref2) {
        var opts = _ref2.opts;

        if (opts != null && opts.names !== undefined) {
          NAMES = Object.assign({}, NAMES, opts.names);
          ENV_NAMES = Object.assign({}, ENV_NAMES, opts.envNames);
          ENV_LABEL_SET = new Set(Object.values(ENV_NAMES));
        } else {
          ENV_LABEL_SET = new Set();
        }

        return path.traverse({
          Function: function Function(fn) {
            if (fn.isArrowFunctionExpression() && !fn.get('body').isBlockStatement()) {
              // Naked arrow functions cannot contain contracts.
              return;
            }

            fn.traverse({
              Function: function Function(path) {
                // This will be handled by the outer visitor, so skip it.
                path.skip();
              },
              LabeledStatement: function LabeledStatement(path) {
                var label = path.get('label');

                if (opts.strip || opts.env && opts.env[process.env.NODE_ENV] && opts.env[process.env.NODE_ENV].strip) {
                  if (label.node.name === NAMES.precondition || label.node.name === NAMES.postcondition || label.node.name === NAMES.invariant || label.node.name === NAMES.assert) {
                    path.remove();
                  }

                  return;
                }

                if (opts.envStrip && ENV_NAMES[process.env.NODE_ENV] !== label.node.name) {
                  path.remove();
                  return;
                }

                var id;
                var children;
                var parent = fn;

                if (label.node.name === NAMES.precondition) {
                  assemblePrecondition(path);
                  return;
                } else if (label.node.name === NAMES.postcondition) {
                  id = assemblePostcondition(path);
                  children = fn.get('body').get('body');
                } else if (label.node.name === NAMES.assert) {
                  assembleAssertion(path);
                  return;
                } else if (label.node.name === NAMES.invariant) {
                  id = assembleInvariant(path);
                  parent = path.findParent(t.isBlockStatement);
                  children = parent.get('body');
                  var first = children[0];
                  first.insertAfter(t.expressionStatement(t.callExpression(id, [])));
                } else if (ENV_LABEL_SET.has(label.node.name)) {
                  extractEnvironmentCode(path);
                  return;
                }

                parent.traverse({
                  Function: function Function(path) {
                    // This will be handled by the outer visitor, so skip it.
                    path.skip();
                  },
                  ReturnStatement: function ReturnStatement(statement) {
                    statement.get('argument').replaceWith(t.callExpression(id, [statement.node.argument]));
                  }
                });
                var last = children[children.length - 1];

                if (!last.isReturnStatement()) {
                  last.insertAfter(t.expressionStatement(t.callExpression(id, [])));
                }
              }
            });
          },
          LabeledStatement: function LabeledStatement(path) {
            var label = path.get('label');

            if (label.node.name === NAMES.assert) {
              if (opts.strip || opts.env && opts.env[process.env.NODE_ENV] && opts.env[process.env.NODE_ENV].strip) {
                path.remove();
              } else {
                assembleAssertion(path);
              }

              return;
            } else if (opts.envStrip) {
              if (ENV_NAMES[process.env.NODE_ENV] !== label.node.name) {
                extractEnvironmentCode(path);
              } else {
                path.remove();
              }

              return;
            }
          }
        });
      }
    }
  };
}